from __future__ import annotations

from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from asyncpg.pool import Pool

    from .creatable import Creatable
    from .types import CustomType


NOT_CREATABLE: list[type[Creatable]] = []

CUSTOM_TYPES: dict[str, type[CustomType[Any]]] = {}

POOLS: list[Pool] = []

DEFAULT_SCHEMA: str = "public"



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, Optional, TYPE_CHECKING, TypeVar, Union

from .types import SQLType
from .utils import MISSING

if TYPE_CHECKING:
    from .selectable import Selectable


__all__ = ("Column",)


T = TypeVar("T")


@dataclass
class Column(Generic[T]):
    """A representation of a SQL Database Table Column

    Parameters
    ----------
        primary_key: :class:`bool`
            Whether the column is a primary key column, defaults to ``False``.
        index: :class:`bool`
            Whether to create an index for the given column, defaults to ``False``.

    Attributes
    ----------
        name: :class:`str`
            The column's name.
        table: :class:`~.Table`
            The Table the column is a part of.
        py_type: Type
            The python type associated with the column.
        sql_type: Type[:class:`~.SQLType`]
            The SQL type associated with the column.
        primary_key: :class:`bool`
            Whether the column is a primary key column.
        index: :class:`bool`
            Whether the column is indexed.
        nullable: :class:`bool`
            Whether the column is nullable.
        unique: :class:`bool`
            Whether the column has a unique constraint.
        default: Any
            The default value of the column.
        references: Optional[:class:`Column`]
            The column which this column references, if set.
    """

    if TYPE_CHECKING:
        name: str
        table: type[Selectable]
        _sql_type: type[SQLType[T]]

    primary_key: bool = False
    index: bool = False
    nullable: bool = True
    unique: bool = False
    default: Optional[Union[str, T]] = MISSING
    references: Optional[Column[T]] = None

    @property
    def py_type(self) -> type[T]:
        return self._sql_type.py_type

    @property
    def sql_type(self) -> type[SQLType[T]]:
        return self._sql_type

    @classmethod
    def _with_type(cls, type: type[SQLType[T]], **options: Any) -> Column[T]:
        column = cls(**options)
        column._sql_type = type
        return column



from __future__ import annotations

from collections.abc import Iterable

from typing import Any, cast, Optional, TYPE_CHECKING, Union, overload

from .column import Column
from .selectable import Selectable
from .utils import not_creatable, query_builder

if TYPE_CHECKING:
    from asyncpg import Connection, Record  # type: ignore


__all__ = ("Insertable",)


@not_creatable
class Insertable(Selectable):

    # region: query generation

    @classmethod
    def _get_columns(
        cls,
        values: dict[str, Any],
    ) -> Iterable[Column]:
        return [cls._columns_dict[column] for column in values]

    @classmethod
    def _get_primary_keys(
        cls,
        record: Record,
    ) -> dict[str, Any]:
        return {column.name: record[column.name] for column in cls._primary_keys}

    @classmethod
    @query_builder
    def _build_query_insert(
        cls,
        columns: Iterable[Column],
        ignore_on_conflict: bool,
        update_on_conflict: Iterable[Column],
        returning: Iterable[Column],
    ) -> list[str]:
        builder = [f"INSERT INTO {cls._name} ("]

        for column in columns:
            builder.append(column.name)
            builder.append(",")

        builder.pop(-1)

        builder.append(") VALUES (")

        for i, _ in enumerate(columns, 1):
            builder.append(f"${i}")
            builder.append(",")

        builder.pop(-1)

        builder.append(")")

        if ignore_on_conflict and update_on_conflict:
            raise ValueError("")

        elif ignore_on_conflict:
            builder.append("ON CONFLICT DO NOTHING")

        elif update_on_conflict:
            builder.append("ON CONFLICT (")

            for column in cls._primary_keys:
                builder.append(column.name)
                builder.append(",")

            builder.pop(-1)

            builder.append(") DO UPDATE SET")

            for column in update_on_conflict:
                builder.append(f"{column.name} = EXCLUDED.{column.name}")
                builder.append(",")

            builder.pop(-1)

        if returning:
            builder.append("RETURNING")

            for column in returning:
                builder.append(column.name)
                builder.append(",")

            builder.pop(-1)

        return builder

    @classmethod
    @query_builder
    def _build_query_update(
        cls,
        where: str,
        offset: int,
        columns: Iterable[Column],
    ) -> list[str]:
        builder = [f"UPDATE", cls._name, "SET"]

        for i, column in enumerate(columns, offset):
            builder.append(column.name)
            builder.append(f"= ${i}")
            builder.append(",")

        builder.pop(-1)

        builder.append("WHERE")
        builder.append(where)

        return builder

    @classmethod
    @query_builder
    def _build_query_delete(
        cls,
        where: str,
    ) -> list[str]:
        builder = ["DELETE FROM", cls._name]
        if where:
            builder.append("WHERE")
            builder.append(where)
        return builder

    # endregion

    # region: public methods

    @overload
    @classmethod
    async def insert(
        cls,
        connection: Connection,
        /,
        *,
        ignore_on_conflict: bool = ...,
        update_on_conflict: Optional[Iterable[Column]] = ...,
        returning: Iterable[Column] = ...,
        **values: Any,
    ) -> Record:
        ...

    @overload
    @classmethod
    async def insert(
        cls,
        connection: Connection,
        *,
        ignore_on_conflict: bool = ...,
        update_on_conflict: Optional[Iterable[Column]] = ...,
        returning: None = ...,
        **values: Any,
    ) -> None:
        ...

    @classmethod
    async def insert(
        cls,
        connection: Connection,
        /,
        *,
        ignore_on_conflict: bool = False,
        update_on_conflict: Optional[Iterable[Column]] = None,
        returning: Optional[Iterable[Column]] = None,
        **values: Any,
    ) -> Optional[Record]:
        r"""|coro|

        Inserts a new record into the database.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        ignore_on_conflict: :class:`bool`
            Sets whether to ignore errors when inserting, defaults to ``False``.
        update_on_conflict: Optional[Iterable[Column]]
            An Optional list of columns to update with new data if a conflict occurs.
        returning: Optional[Iterable[:class:`~.Column`]]
            An optional list of values to return from the inserted record.
        \*\*values: Any
            The column to value mapping for the record to insert.

        Returns
        -------
        Optional[:class:`asyncpg.Record`]
            A record containing information from the inserted record.
        """
        columns = cls._get_columns(values)
        query = cls._build_query_insert(columns, ignore_on_conflict, update_on_conflict, returning or [])
        if returning is not None:
            return await connection.fetchrow(query, *values.values())
        await connection.execute(query, *values.values())

    @overload
    @classmethod
    async def insert_many(
        cls,
        connection: Connection,
        /,
        columns: Iterable[Column],
        *values: Iterable[Any],
        ignore_on_conflict: bool = False,
        update_on_conflict: Optional[Iterable[Column]] = None,
    ) -> None:
        ...

    @overload
    @classmethod
    async def insert_many(
        cls,
        connection: Connection,
        /,
        columns: None,
        *values: dict[str, Any],
        ignore_on_conflict: bool = False,
        update_on_conflict: Optional[Iterable[Column]] = None,
    ) -> None:
        ...

    @classmethod
    async def insert_many(
        cls,
        connection: Connection,
        /,
        columns: Optional[Iterable[Column]],
        *values: Union[Iterable[Any], dict[str, Any]],
        ignore_on_conflict: bool = False,
        update_on_conflict: Optional[Iterable[Column]] = None,
    ) -> None:
        r"""|coro|

        Inserts a set of new records into the database.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        \*values: Dict[:class:`str`, Any]
            The column to value mappings for each record to insert.
        ignore_on_conflict: :class:`bool`
            Sets whether to ignore errors when inserting, defaults to ``False``.
        update_on_conflict: Optional[Iterable[Column]]
            An Optional list of columns to update with new data if a conflict occurs.
        """
        if columns is None:
            values = cast(tuple[dict[str, Any], ...], values)
            columns = cls._get_columns(values[0])
            values = cast(tuple[Iterable[Any]], (value.values() for value in values))

        query = cls._build_query_insert(columns, ignore_on_conflict, update_on_conflict or [], [])
        await connection.executemany(query, values)

    @classmethod
    async def update_where(
        cls,
        connection: Connection,
        /,
        where: str,
        *values: Any,
        **_values: Any,
    ) -> None:
        r"""|coro|

        Updates records in the database which match a given WHERE clause.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        where: :class:`str`
            An SQL WHERE clause.
        \*values: Any
            Values to be substituted into the WHERE clause.
        \*\*values: Any
            The column to value mapping to assign to updated records.
        """
        columns = cls._get_columns(_values)
        query = cls._build_query_update(where, len(_values) + 1, columns)
        await connection.execute(query, *values, *_values.values())

    @classmethod
    async def update_record(
        cls,
        connection: Connection,
        /,
        record: Record,
        **values: Any,
    ) -> None:
        r"""|coro|

        Updates a record in the database.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        record: :class:`asyncpg.Record`
            The record to update.
        \*\*values: Any
            The column to value mapping to assign to updated record.
        """
        primary_keys = cls._get_primary_keys(record)
        where = cls._build_where_clause(primary_keys)
        return await cls.update_where(connection, where, *primary_keys.values(), **values)

    @classmethod
    async def delete_where(
        cls,
        connection: Connection,
        /,
        where: str,
        *values: Any,
    ) -> None:
        """|coro|

        Deletes records in the database which match the given WHERE clause.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        where: :class:`str`
            An SQL WHERE clause.
        *values: Any
            Values to be substituted into the WHERE clause.
        """
        query = cls._build_query_delete(where)
        await connection.execute(query, *values)

    @classmethod
    async def delete(
        cls,
        connection: Connection,
        /,
        **values: Any,
    ) -> None:
        r"""|coro|

        Deletes records in the database which contain the given values.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        \*\*values: Any
            The column to value mapping to filter records with.
        """
        where = cls._build_where_clause(values)
        return await cls.delete_where(connection, where, *values.values())

    @classmethod
    async def delete_record(
        cls,
        connection: Connection,
        /,
        record: Record,
    ) -> None:
        """|coro|

        Deletes a given record from the database.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        record: :class:`asyncpg.Record`
            The record to delete.
        """
        primary_keys = cls._get_primary_keys(record)
        where = cls._build_where_clause(primary_keys)
        return await cls.delete_where(connection, where, *primary_keys.values())

    # endregion



from __future__ import annotations

from collections.abc import Callable
from typing import Any, TYPE_CHECKING

from .insertable import Insertable
from .utils import MISSING, not_creatable, query_builder

if TYPE_CHECKING:
    from asyncpg import Connection, Record  # type: ignore


__all__ = ("Table",)


@not_creatable
class Table(Insertable):
    """Base class for creating representations of SQL Database Tables.

    Attributes
    ----------
        _name: :class:`str`
            The name of the table.
        _schema: :class:`str`
            The tables schema.
        _columns: Iterable[:class:`~.Column`]
            The columns contained in the table.
        _columns_dict: Dict[:class:`str`, :class:`~.Column`]
            A mapping between a column's name and itself.
        _primary_keys: Iterable[:class:`~.Column`]
            The primary key columns of the table.
    """

    @classmethod
    @query_builder
    def _query_create(cls, if_not_exists: bool) -> list[str]:
        builder = ["CREATE TABLE"]

        if if_not_exists:
            builder.append("IF NOT EXISTS")

        builder.append(cls._name)
        builder.append("(")

        for column in cls._columns:
            builder.append(column.name)

            builder.append(column.sql_type.sql_type)

            if not column.nullable:
                builder.append("NOT NULL")

            if column.unique:
                builder.append("UNIQUE")

            if column.default is not MISSING:
                builder.append("DEFAULT")
                builder.append(str(column.default))

            if column.references is not None:
                builder.append("REFERENCES")
                builder.append(column.references.table._name)
                builder.append("(")
                builder.append(column.references.name)
                builder.append(")")

            builder.append(",")

        if cls._primary_keys:
            builder.append("PRIMARY KEY (")
            for column in cls._primary_keys:
                builder.append(column.name)
                builder.append(",")

            builder.pop(-1)
            builder.append(")")
        else:
            builder.pop(-1)

        builder.append(")")

        return builder

    @classmethod
    def _query_drop(cls, if_exists: bool, cascade: bool) -> str:
        return super()._query_drop("TABLE", if_exists, cascade)

    @classmethod
    async def migrate_to(
        cls,
        connection: Connection,
        /,
        table: type[Table],
        migration: Callable[[Record], dict[str, Any]],
        *,
        create_new_table: bool = False,
        drop_table: bool = False,
    ) -> None:
        """|coro|

        Helper function for migrating data in a table to another.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        table: Type[:class:`.Table`]
            The new table to migrate to.
        migration: Callable[[:class:`asyncpg.Record`], Dict[:class:`str`, Any]]
            The function used to migrate data between tables.
        create_new_table: :class:`bool`
            Sets whether the table to migrate to should be created.
            Defaults to ``False``.
        drop_table: :class:`bool`
            Sets whether this table should be dropped after migrating.
            Defaults to ``False``.
        """

        if create_new_table:
            await table.create(connection)

        records = await cls.fetch(connection)
        await table.insert_many(connection, None, *(migration(record) for record in records))

        if drop_table:
            await cls.drop(connection)



from __future__ import annotations

import datetime
import json

from collections.abc import Callable
from typing import Any, TYPE_CHECKING, NamedTuple, TypeVar, Literal

import asyncpg

from .consts import CUSTOM_TYPES, POOLS

if TYPE_CHECKING:
    from asyncpg import Connection
    from asyncpg.pool import Pool

__all__ = (
    "create_pool",
    "MaybeAcquire",
    "TYPE_CODECS",
    "OPTIONAL_CODECS",
)


T = TypeVar("T")


# Y2K_DT = datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc)
Y2K_EPOCH = 946684800000000


class TypeCodec(NamedTuple):
    format: Literal["text", "binary", "tuple"]
    encoder: Callable[..., Any]
    decoder: Callable[..., Any]


def _encode_datetime(value: datetime.datetime) -> tuple[int]:
    if value.tzinfo is None:
        value = value.astimezone(datetime.timezone.utc)
    return (round(value.timestamp() * 1_000_000) - Y2K_EPOCH,)


def _decode_timestamp(value: tuple[int]) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(round(value[0] + Y2K_EPOCH) / 1_000_000, datetime.timezone.utc)


TYPE_CODECS: dict[str, TypeCodec] = {
    "json": TypeCodec("text", json.dumps, json.loads),
    "jsonb": TypeCodec("text", json.dumps, json.loads),
}

OPTIONAL_CODECS: dict[str, TypeCodec] = {
    "timestamp": TypeCodec("tuple", _encode_datetime, _decode_timestamp),
}


async def create_pool(dsn: str, codecs: dict[str, TypeCodec] = {}, **kwargs) -> Pool:
    r"""|coro|

    Creates the database connection pool.

    Parameters
    ----------
    dsn: :class:`str`
        A database connection string.
    codecs: Dict[:class:`str`, :class:`.TypeCodec`]
        A mapping of type to encoder and decoder for custom type codecs.
    \*\*kwargs: Any
        Extra keyword arguments to pass to :meth:`asyncpg.create_pool`

    Returns
    -------
    :class:`asyncpg.pool.Pool`
        The new pool which was created.
    """
    codecs |= TYPE_CODECS

    async def init(connection: Connection) -> None:
        for type, codec in codecs.items():
            await connection.set_type_codec(
                type, schema="pg_catalog", encoder=codec.encoder, decoder=codec.decoder, format=codec.format
            )

        for type in CUSTOM_TYPES.values():
            await type._set_codec(connection)

    pool = await asyncpg.create_pool(dsn, init=init, **kwargs)
    if pool is None:
        raise RuntimeError("Could not create pool.")
    POOLS.append(pool)
    return pool


class MaybeAcquire:
    """Async helper for acquiring a connection to the database.

    .. container:: operations

        .. describe:: async with x as c

            Yields a database connection to use,
            if a new one was created it will be closed on exit.

    Parameters
    ----------
        connection: Optional[:class:`asyncpg.Connection <asyncpg.connection.Connection>`]
            A database connection to use.
            If none is supplied a connection will be acquired from the pool.
        pool: Optional[:class:`asyncpg.pool.Pool`]
            A connection pool to use.

    Attributes
    ----------
        connection: Optional[:class:`asyncpg.Connection <asyncpg.connection.Connection>`]
            The supplied database connection, if provided.
        pool: Optional[:class:`asyncpg.pool.Pool`]
            The connection pool used to acquire new connections.
    """

    def __init__(self, connection: Connection = None, /, *, pool: Pool):
        self.connection = connection
        self.pool = pool
        self._cleanup = False

    async def __aenter__(self) -> Connection:
        if self.connection is None:
            self._cleanup = True
            self._connection = c = await self.pool.acquire()
            return c
        return self.connection

    async def __aexit__(self, *args):
        if self._cleanup:
            await self.pool.release(self._connection)



from __future__ import annotations

import types

from typing import Any, NamedTuple, TYPE_CHECKING, TypeVar

__all__ = ("Enum",)


class _EnumValue(NamedTuple):
    name: str
    value: Any


if TYPE_CHECKING:
    from enum import Enum as _Enum

    class Enum(_Enum):
        if TYPE_CHECKING:

            @classmethod
            def try_value(cls: type[ET], value: Any) -> ET:
                ...


else:

    def _create_value_cls(name):
        cls = types.new_class(f"_EnumValue_{name}", (_EnumValue,))
        cls.__repr__ = lambda self: f"<{name}.{self.name}: {self.value!r}>"
        cls.__str__ = lambda self: f"{name}.{self.name}"
        return cls

    def _is_descriptor(obj):
        return hasattr(obj, "__get__") or hasattr(obj, "__set__") or hasattr(obj, "__delete__")

    class EnumMeta(type):
        def __new__(cls, name, bases, attrs):
            value_mapping = {}
            member_mapping = {}
            member_names = []

            value_cls = _create_value_cls(name)
            for key, value in list(attrs.items()):
                is_descriptor = _is_descriptor(value)
                if key[0] == "_" and not is_descriptor:
                    continue

                # Special case classmethod to just pass through
                if isinstance(value, classmethod):
                    continue

                if is_descriptor:
                    setattr(value_cls, key, value)
                    del attrs[key]
                    continue

                try:
                    new_value = value_mapping[value]
                except KeyError:
                    new_value = value_cls(name=key, value=value)
                    value_mapping[value] = new_value
                    member_names.append(key)

                member_mapping[key] = new_value
                attrs[key] = new_value

            attrs["_enum_value_map_"] = value_mapping
            attrs["_enum_member_map_"] = member_mapping
            attrs["_enum_member_names_"] = member_names
            attrs["_enum_value_cls_"] = value_cls
            actual_cls = super().__new__(cls, name, bases, attrs)
            value_cls._actual_enum_cls_ = actual_cls
            return actual_cls

        def __iter__(cls):
            return (cls._enum_member_map_[name] for name in cls._enum_member_names_)

        def __reversed__(cls):
            return (cls._enum_member_map_[name] for name in reversed(cls._enum_member_names_))

        def __len__(cls):
            return len(cls._enum_member_names_)

        def __repr__(cls):
            return f"<enum {cls.__name__}>"

        @property
        def __members__(cls):
            return types.MappingProxyType(cls._enum_member_map_)

        def __call__(cls, value):
            try:
                return cls._enum_value_map_[value]
            except (KeyError, TypeError):
                raise ValueError(f"{value!r} is not a valid {cls.__name__}")

        def __getitem__(cls, key):
            return cls._enum_member_map_[key]

        def __setattr__(cls, name, value):
            raise TypeError("Enums are immutable.")

        def __delattr__(cls, attr):
            raise TypeError("Enums are immutable")

        def __instancecheck__(self, instance):
            # isinstance(x, Y)
            # -> __instancecheck__(Y, x)
            try:
                return instance._actual_enum_cls_ is self
            except AttributeError:
                return False

    class Enum(metaclass=EnumMeta):
        @classmethod
        def try_value(cls, value):
            try:
                return cls._enum_value_map_[value]
            except (KeyError, TypeError):
                return value


ET = TypeVar("ET", bound=Enum)


def create_unknown_value(cls: type[ET], val: Any) -> ET:
    value_cls = cls._enum_value_cls_  # type: ignore
    name = f"unknown_{val}"
    return value_cls(name=name, value=val)


def try_enum(cls: type[ET], val: Any) -> ET:
    """A function that tries to turn the value into enum ``cls``.
    If it fails it returns a proxy invalid value instead.
    """

    try:
        return cls._enum_value_map_[val]  # type: ignore
    except (KeyError, TypeError, AttributeError):
        return create_unknown_value(cls, val)



from __future__ import annotations

import decimal
import datetime
import ipaddress
import uuid

from types import new_class
from typing import Any, Generic, NamedTuple, Optional, TYPE_CHECKING, TypeVar, Union

import asyncpg

from .consts import CUSTOM_TYPES, POOLS
from .creatable import Creatable
from .enums import Enum
from .utils import BUILDING_DOCS, normalise_name, not_creatable, query_builder

if TYPE_CHECKING:
    from asyncpg import Connection


__all__ = (
    "SQLType",
    "CustomType",
    "EnumType",
)


T = TypeVar("T")
OT = TypeVar("OT")
ET = TypeVar("ET", bound=Enum)


class SQLType(Generic[T]):
    """A representation of an SQL type

    Attributes
    ----------
        py_type: Type[Any]
            The python type associated with the column.
        sql_type: Type[:class:`~.SQLType`]
            The SQL type associated with the column.
    """

    if TYPE_CHECKING:
        # 8.1 Numeric
        Integer = Int = int
        SmallInt = int
        BigInt = int
        Serial = int
        float = float
        dounle = float
        numeric = decimal.Decimal
        # 8.2 Monetary
        Money = str
        # 8.3 Character
        # CharacterVarying = VarChar = str
        Character = str
        Text = str
        # 8.4 Binary
        Bytea = bytes
        # 8.5 Date/Time
        Timestamp = datetime.datetime
        TimestampAware = datetime.datetime
        Date = datetime.date
        Inteval = datetime.timedelta
        # 8.6 Boolean
        Boolean = bool
        # 8.9 Network Adress
        CIDR = ipaddress._BaseNetwork
        Inet = ipaddress._BaseNetwork
        MACAddr = str
        # 8.12 UUID
        UUID = uuid.UUID
        # 8.14 JSON
        JSON = dict
        JSONB = dict

    py_type: type[T]
    sql_type: str
    __defaults: dict[type[Any], type[SQLType[Any]]] = {}
    __enum_types: dict[type[Any], type[SQLType[Any]]] = {}

    def __init_subclass__(
        cls,
        *,
        sql_type: Optional[str] = None,
        default: bool = False,
        **kwargs: Any,
    ) -> None:
        cls.py_type = cls.__orig_bases__[0].__args__[0]  # type: ignore
        if default:
            cls.__defaults[cls.py_type] = cls
        super().__init_subclass__(**kwargs)
        cls.sql_type = sql_type or cls._name  # type: ignore

    @classmethod
    def _from_type(cls, type: type[OT]) -> type[SQLType[OT]]:
        if issubclass(type, Enum):
            if type in cls.__enum_types:
                return cls.__enum_types[type]

            enum_type = new_class(type.__name__, (EnumType[type],))
            cls.__enum_types[type] = enum_type
            return enum_type

        return cls.__defaults[type]


@not_creatable
class CustomType(SQLType[T], Creatable, sql_type=""):
    """A representation of a Custom SQL type.

    Attributes
    ----------
        py_type: Type[Any]
            The python type associated with the column.
        sql_type: Type[:class:`~.SQLType`]
            The SQL type associated with the column.
        _name: :class:`str`
            The name of the table.
        _schema: :class:`str`
            The tables schema.
    """

    @classmethod
    def _query_drop(cls, if_exists: bool, cascade: bool) -> str:
        return super()._query_drop("TYPE", if_exists, cascade)

    @classmethod
    async def _set_codec(cls, connection: Connection) -> None:
        raise NotImplementedError()

    @classmethod
    async def create(cls, connection: Connection, /, *args: Any, if_not_exists: bool = True, **kwargs: Any) -> None:
        try:
            await super().create(connection, *args, **kwargs)
        except asyncpg.exceptions.DuplicateObjectError:
            if not if_not_exists:
                raise
        CUSTOM_TYPES[cls._name] = cls

        for pool in POOLS:
            for holder in pool._holders:
                await cls._set_codec(holder._con)

    @classmethod
    async def drop(cls, connection: Connection, /, *args: Any, **kwargs: Any) -> None:
        for pool in POOLS:
            for holder in pool._holders:
                await holder._con.reset_type_codec(cls._name[len(cls._schema) + 1 :], schema=cls._schema)

        await super().drop(connection, *args, **kwargs)
        del CUSTOM_TYPES[cls._name]


@not_creatable
class EnumType(CustomType[ET], sql_type=""):
    """A representations of an SQL Enum type.

    Attributes
    ----------
        py_type: Type[:class:`~.Enum`]
            The python type associated with the column.
        sql_type: Type[:class:`~.SQLType`]
            The SQL type associated with the column.
        _name: :class:`str`
            The name of the table.
        _schema: :class:`str`
            The tables schema.
    """

    @classmethod
    @query_builder
    def _query_create(cls, if_not_exists: bool) -> list[str]:
        builder = ["CREATE TYPE"]
        builder.append(cls._name)
        builder.append("AS ENUM (")

        for key in cls.py_type:  # type: ignore
            builder.append(f"'{key.name}'")
            builder.append(",")

        builder.pop(-1)

        builder.append(")")
        return builder

    @classmethod
    def _encoder(cls, value: Union[ET, Any]) -> str:
        return cls.py_type.try_value(value).name

    @classmethod
    def _decoder(cls, value: str) -> ET:
        return getattr(cls.py_type, value)

    @classmethod
    async def _set_codec(cls, connection: Connection) -> None:
        await connection.set_type_codec(
            cls._name[len(cls._schema) + 1 :],
            schema=cls._schema,
            encoder=cls._encoder,
            decoder=cls._decoder,
            format="text",
        )


if not TYPE_CHECKING:

    class SQLTypeConfig(NamedTuple):
        py_type: type[Any]
        sql_type: str
        is_default: bool = False
        aliases: tuple[str, ...] = ()

    for name, (py_type, sql_type, is_default, aliases) in {
        # 8.1 Numeric
        "Integer": SQLTypeConfig(int, "INTEGER", True, ("Int",)),
        "SmallInt": SQLTypeConfig(int, "SMALLINT"),
        "BigInt": SQLTypeConfig(int, "BIGINT"),
        "Serial": SQLTypeConfig(int, "SERIAL"),
        "Float": SQLTypeConfig(float, "FLOAT", True),
        "DoublePrecision": SQLTypeConfig(float, "DOUBLE PRECISION", False, ("Double",)),
        "Numeric": SQLTypeConfig(decimal.Decimal, "NUMERIC", True),
        # 8.2 Monetary
        "Money": SQLTypeConfig(str, "MONEY"),
        # 8.3 Character
        # "CharacterVarying": SQLTypeConfig(str, "CHARACTER VARYING({n})", False, ("VarChar",)),
        "Character": SQLTypeConfig(str, "CHARACTER", False, ("Char",)),
        "Text": SQLTypeConfig(str, "TEXT", True),
        # 8.4 Binary
        "Bytea": SQLTypeConfig(bytes, "BYTEA", True),
        # 8.5 Date/Time
        "Timestamp": SQLTypeConfig(datetime.datetime, "TIMESTAMP", True, ("NaieveTimestamp", "Datetime")),
        "AwareTimestamp": SQLTypeConfig(datetime.datetime, "TIMESTAMP WITH TIME ZONE", True),
        "Date": SQLTypeConfig(datetime.date, "DATE", True),
        "Interval": SQLTypeConfig(datetime.timedelta, "INTERVAL", True),
        # 8.6 Boolean
        "Boolean": SQLTypeConfig(bool, "BOOLEAN", True),
        # 8.9 Network Adress
        "CIDR": SQLTypeConfig(ipaddress._BaseNetwork, "CIDR"),
        "Inet": SQLTypeConfig(ipaddress._BaseNetwork, "INET"),
        "MACAddr": SQLTypeConfig(str, "MACADDR"),
        # 8.12 UUID
        "UUID": SQLTypeConfig(uuid.UUID, "UUID", True),
        # 8.14 JSON
        "JSON": SQLTypeConfig(dict, "JSON"),
        "JSONB": SQLTypeConfig(dict, "JSONB", True),
    }.items():
        cls = new_class(name, (SQLType[py_type],), {"sql_type": sql_type, "default": is_default})
        new_class(name + "[]", (SQLType[list[py_type]],), {"sql_type": sql_type + "[]", "default": is_default})

        if BUILDING_DOCS:

            @property
            def _(cls):
                return cls

            _.__doc__ = f"Represents the SQL ``{sql_type}`` type."
            if is_default:
                qualified_name = ""
                if py_type.__module__ != "builtins":
                    qualified_name = f"{py_type.__module__}."
                qualified_name += py_type.__name__
                _.__doc__ += f" Python class :class:`{qualified_name}`, can be used as a substitute."

        else:
            _ = cls

        setattr(SQLType, name, _)
        for alias in aliases:
            setattr(SQLType, alias, _)



from __future__ import annotations

from typing import Protocol, TYPE_CHECKING

from .consts import DEFAULT_SCHEMA, NOT_CREATABLE
from .utils import MISSING, normalise_name, query_builder

if TYPE_CHECKING:
    from asyncpg import Connection


__all__ = ("Creatable",)


class Creatable(Protocol):
    _schema: str
    _name: str

    def __init_subclass__(
        cls,
        *,
        _name: str = MISSING,
        schema: str = MISSING,
    ) -> None:
        if schema is MISSING:
            schema = DEFAULT_SCHEMA

        if _name is MISSING:
            _name = normalise_name(cls.__name__)

        cls._schema = schema
        cls._name = f"{schema}.{_name}"
        super().__init_subclass__()

    # region: query generation

    @classmethod
    def _query_create(
        cls,
        if_not_exists: bool,
    ) -> str:
        raise NotImplementedError()

    @classmethod
    @query_builder
    def _query_create_schema(
        cls,
        if_not_exists: bool,
    ) -> list[str]:
        builder = ["CREATE SCHEMA"]
        if if_not_exists:
            builder.append("IF NOT EXISTS")
        builder.append(cls._schema)
        return builder

    @classmethod
    @query_builder
    def _query_drop(
        cls,
        type: str,
        if_exists: bool,
        cascade: bool,
    ) -> list[str]:
        builder = ["DROP", type]

        if if_exists:
            builder.append("IF EXISTS")

        builder.append(cls._name)

        if cascade:
            builder.append("CASCADE")

        return builder

    # endregion

    # region: public methods

    @classmethod
    async def create(
        cls,
        connection: Connection,
        *,
        if_not_exists: bool = True,
        create_schema: bool = True,
    ) -> None:
        """|coro|

        Creates this database object.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        if_not_exists: :class:`bool`
            Sets whether creation should continue if the object already exists.
            Defaults to ``True``.
        create_schema: :class:`bool`
            Sets whether the database schema should also be created.
            Defaults to ``True``.
        """
        if create_schema:
            await cls.create_schema(connection, if_not_exists=if_not_exists)
        query = cls._query_create(if_not_exists)
        await connection.execute(query)

    @classmethod
    async def create_schema(
        cls,
        connection: Connection,
        *,
        if_not_exists: bool = True,
    ) -> None:
        """|coro|

        Creates the schema this database object uses.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        if_not_exists: :class:`bool`
            Sets whether creation should continue if the schema already exists.
            Defaults to ``True``.
        """
        query = cls._query_create_schema(if_not_exists)
        await connection.execute(query)

    @classmethod
    async def create_all(
        cls,
        connection: Connection,
        /,
        *,
        if_not_exists: bool = True,
        create_schema: bool = True,
    ) -> None:
        """|coro|

        Creates all subclasses of this database object.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        if_not_exists: :class:`bool`
            Sets whether creation should continue if the object already exists.
            Defaults to ``True``.
        create_schema: :class:`bool`
            Sets whether the database schema should also be created.
            Defaults to ``True``.
        """
        for subcls in cls.__subclasses__():
            if subcls in NOT_CREATABLE:
                await subcls.create_all(
                    connection,
                    if_not_exists=if_not_exists,
                    create_schema=create_schema,
                )
            else:
                await subcls.create(
                    connection,
                    if_not_exists=if_not_exists,
                    create_schema=create_schema,
                )

    @classmethod
    async def drop(
        cls,
        connection: Connection,
        /,
        *,
        if_exists: bool = True,
        cascade: bool = False,
    ) -> None:
        """|coro|

        Drops this object from the database.

        Parameters
        ----------
        if_exists: :class:`bool`
            Sets whether dropping should not error if the object does not exist.
            Defaults to ``True``.
        cascade: :class:`bool`
            Sets whether dropping should cascade.
            Defaults to ``False``.
        """
        query = cls._query_drop(if_exists, cascade)
        await connection.execute(query)

    # endregion



from __future__ import annotations

import inspect
import sys

from collections.abc import Iterable
from typing import Any, Literal, Optional, TYPE_CHECKING

from .column import Column, SQLType
from .creatable import Creatable
from .utils import not_creatable, query_builder, resolve_annotation

if TYPE_CHECKING:
    from asyncpg import Connection, Record  # type: ignore


__all__ = ("Selectable",)


OPERATORS: dict[str, str] = {
    "eq": "=",
    "lt": "<",
    "le": "<=",
    "ne": "<>",
    "ge": ">=",
    "gt": ">",
}


OrderBy = tuple[Column, Literal["ASC", "DESC"]]


@not_creatable
class Selectable(Creatable):
    if TYPE_CHECKING:
        _columns: Iterable[Column]
        _columns_dict: dict[str, Column]
        _primary_keys: Iterable[Column]

    def __init_subclass__(cls, **kwargs: Any) -> None:
        cache: dict[str, Any] = {}
        cls._columns = []
        cls._columns_dict = {}
        cls._primary_keys = []

        # Get global namespaces
        try:
            globals = sys.modules[cls.__module__].__dict__
        except KeyError:
            globals = {}

        # Get local namespace
        frame = inspect.currentframe()
        try:
            if frame is None:
                locals = {}
            else:
                if frame.f_back is None:
                    locals = frame.f_locals
                else:
                    locals = frame.f_back.f_locals
        finally:
            del frame

        for name, type in cls.__annotations__.items():
            type = resolve_annotation(type, globals, locals, cache)

            if getattr(type, "__origin__", None) is not Column:
                raise TypeError("Column typings must be of type Column.")

            type = type.__args__[0]
            is_array = False

            if getattr(type, "__origin__", None) is list:
                is_array = True
                type = type.__args__[0]

            try:
                if not issubclass(type, SQLType):
                    type = SQLType._from_type(list[type] if is_array else type)  # type: ignore
                elif is_array:
                    type = SQLType._from_type(list[type.py_type])  # type: ignore

            except TypeError:
                if getattr(type, "__origin__", None) is not SQLType:
                    raise TypeError("Column typing generics must be a valid SQLType.")
                type = type.__args__[0]  # type: ignore

                type = SQLType._from_type(list[type] if is_array else type)  # type: ignore

            if not hasattr(cls, name):
                column = Column._with_type(type)  # type: ignore
                setattr(cls, name, column)
            else:
                column = getattr(cls, name)
                if not isinstance(column, Column):
                    raise ValueError("Column values must be an instance of Column.")

                column._sql_type = type

            column.name = name
            column.table = cls

            cls._columns.append(column)
            cls._columns_dict[name] = column
            if column.primary_key:
                cls._primary_keys.append(column)

        super().__init_subclass__(**kwargs)

    # region: query generation

    @classmethod
    @query_builder
    def _build_where_clause(
        cls,
        values: dict[str, Any],
    ) -> list[str]:
        builder = []

        if len(values) > 1:
            builder.append("(" * (len(values) - 1))

        for i, name in enumerate(values, 1):
            is_or = False
            operator = OPERATORS["eq"]
            if name.startswith("or_"):
                if i == 1:
                    raise NameError("")
                is_or = True
                name = name[3:]

            if i > 1:
                builder.append("OR" if is_or else "AND")

            if name[-4:-2] == "__":
                name, operator = name.rsplit("__", 1)
                if operator not in OPERATORS:
                    raise NameError(f"Unknown operator {operator}.")
                operator = OPERATORS[operator]

            if name not in cls._columns_dict:
                raise NameError(f"Unknown column {name} in table {cls._name}.")

            builder.append(name)
            builder.append(operator)

            builder.append(f"${i}")

            if i > 1:
                builder.append(")")

        return builder

    @classmethod
    @query_builder
    def _build_query_fetch(
        cls,
        where: str,
        limit: Optional[int],
        order_by: Optional[OrderBy],
    ) -> list[str]:
        builder = ["SELECT * FROM", cls._name]
        if where:
            builder.append("WHERE")
            builder.append(where)

        if order_by is not None:
            column, direction = order_by
            builder.append("ORDER BY")
            builder.append(column.name)
            builder.append(direction)

        if limit is not None:
            builder.append("LIMIT")
            builder.append(str(limit))

        return builder

    # endregion

    # region: public methods

    @classmethod
    async def fetch_where(
        cls,
        connection: Connection,
        /,
        where: str,
        *values: Any,
        limit: Optional[int] = None,
        order_by: Optional[OrderBy] = None,
    ) -> Iterable[Record]:
        r"""|coro|

        Fetches records in the database which match a given WHERE clause.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        where: :class:`str`
            An SQL WHERE clause.
        \*values: Any
            Values to be substituted into the WHERE clause.
        limit: Optional[:class:`int`]
            If provided, sets the maxmimum number of records to be returned.
        order_by: Optional[Tuple[:class:`Column`, Literal[`"ASC"``, ``"DESC"``]]]
            Sets the ORDER BY condition on the database query, takes a tuple
            concisting of the column and direction.

        Returns
        -------
        Iterable[:class:`asyncpg.Record`]
            Records which match the WHERE clause.
        """
        query = cls._build_query_fetch(where, limit, order_by)
        return await connection.fetch(query, *values)

    @classmethod
    async def fetch(
        cls,
        connection: Connection,
        /,
        *,
        limit: Optional[int] = None,
        order_by: Optional[OrderBy] = None,
        **values: Any,
    ) -> Iterable[Record]:
        r"""|coro|

        Fetches records in the database which contain the given values.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        limit: Optional[:class:`int`]
            If provided, sets the maxmimum number of records to be returned.
        order_by: Optional[Tuple[:class:`Column`, Literal["ASC", "DESC"]]]
            Sets the ORDER BY condition on the database query, takes a tuple
            concisting of the column and direction.
        \*\*values: Any
            The column to value mapping to filter records with.

        Returns
        -------
        Iterable[:class:`asyncpg.Record`]
            Records which which contain the given values.
        """
        where = cls._build_where_clause(values)
        return await cls.fetch_where(connection, where, *values.values(), limit=limit, order_by=order_by)

    @classmethod
    async def fetch_row_where(
        cls,
        connection: Connection,
        /,
        where: str,
        *values: Any,
        order_by: Optional[OrderBy] = None,
    ) -> Optional[Record]:
        r"""|coro|

        Fetches a record in the database which match a given WHERE clause.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        where: :class:`str`
            An SQL WHERE clause.
        \*values: Any
            Values to be substituted into the WHERE clause.
        order_by: Optional[Tuple[:class:`Column`, Literal["ASC", "DESC"]]]
            Sets the ORDER BY condition on the database query, takes a tuple
            concisting of the column and direction.

        Returns
        -------
        Optional[:class:`asyncpg.Record`]
            A record which matches the WHERE clause if found.
        """
        query = cls._build_query_fetch(where, None, order_by)
        return await connection.fetchrow(query, *values)

    @classmethod
    async def fetch_row(
        cls,
        connection: Connection,
        /,
        *,
        order_by: Optional[OrderBy] = None,
        **values: Any,
    ) -> Optional[Record]:
        r"""|coro|

        Fetches a records in the database which contains the given values.

        Parameters
        ----------
        connection: :class:`asyncpg.Connection <asyncpg.connection.Connection>`
            The database connection to use for transactions.
        order_by: Optional[Tuple[:class:`Column`, Literal["ASC", "DESC"]]]
            Sets the ORDER BY condition on the database query, takes a tuple
            concisting of the column and direction.
        \*\*values: Any
            The column to value mapping to filter records with.

        Returns
        -------
        Optional[:class:`asyncpg.Record`]
            A record which contains the given values if found.
        """
        where = cls._build_where_clause(values)
        return await cls.fetch_row_where(connection, where, *values.values(), order_by=order_by)

    # endregion



from __future__ import annotations

import sys
import types
import string
from functools import wraps

from collections.abc import Callable, Iterable
from typing import Any, ForwardRef, Optional, Literal, Union, TypeVar, TYPE_CHECKING

from .consts import NOT_CREATABLE

if TYPE_CHECKING:
    from .creatable import Creatable


__all__ = (
    "MISSING",
    "normalise_name",
    "query_builder",
    "not_creatable",
    "resolve_annotation",
    "BUILDING_DOCS",
)


T = TypeVar("T")
CT = TypeVar("CT", bound="Creatable")


class _MissingSentinel:
    def __repr__(self):
        return "MISSING"


MISSING: Any = _MissingSentinel()


_normalisation = str.maketrans({u: f"_{l}" for u, l in zip(string.ascii_uppercase, string.ascii_lowercase)})


def normalise_name(name: str) -> str:
    return name[0].lower() + name[1:].translate(_normalisation)


def query_builder(func: Callable[..., list[Any]]) -> Callable[..., str]:
    @wraps(func)
    def wrapper(*args, **kwargs):
        return " ".join(str(e) for e in func(*args, **kwargs))

    return wrapper


def not_creatable(cls: type[CT]) -> type[CT]:
    NOT_CREATABLE.append(cls)
    return cls


PY_310 = sys.version_info >= (3, 10)


def flatten_literal_params(parameters: Iterable[Any]) -> tuple[Any, ...]:
    params = []
    literal_cls = type(Literal[0])
    for p in parameters:
        if isinstance(p, literal_cls):
            params.extend(p.__args__)
        else:
            params.append(p)
    return tuple(params)


def normalise_optional_params(parameters: Iterable[Any]) -> tuple[Any, ...]:
    none_cls = type(None)
    return tuple(p for p in parameters if p is not none_cls) + (none_cls,)


def evaluate_annotation(
    tp: Any,
    globals: dict[str, Any],
    locals: dict[str, Any],
    cache: dict[str, Any],
    *,
    implicit_str: bool = True,
):
    if isinstance(tp, ForwardRef):
        tp = tp.__forward_arg__
        # ForwardRefs always evaluate their internals
        implicit_str = True

    if implicit_str and isinstance(tp, str):
        if tp in cache:
            return cache[tp]
        evaluated = eval(tp, globals, locals)
        cache[tp] = evaluated
        return evaluate_annotation(evaluated, globals, locals, cache)

    if hasattr(tp, "__args__"):
        implicit_str = True
        is_literal = False
        args = tp.__args__
        if not hasattr(tp, "__origin__"):
            if PY_310 and tp.__class__ is types.Union:  # type: ignore
                converted = Union[args]  # type: ignore
                return evaluate_annotation(converted, globals, locals, cache)

            return tp
        if tp.__origin__ is Union:
            try:
                if args.index(type(None)) != len(args) - 1:
                    args = normalise_optional_params(tp.__args__)
            except ValueError:
                pass
        if tp.__origin__ is Literal:
            if not PY_310:
                args = flatten_literal_params(tp.__args__)
            implicit_str = False
            is_literal = True

        evaluated_args = tuple(
            evaluate_annotation(arg, globals, locals, cache, implicit_str=implicit_str) for arg in args
        )

        if is_literal and not all(isinstance(x, (str, int, bool, type(None))) for x in evaluated_args):
            raise TypeError("Literal arguments must be of type str, int, bool, or NoneType.")

        if evaluated_args == args:
            return tp

        try:
            return tp.copy_with(evaluated_args)
        except AttributeError:
            return tp.__origin__[evaluated_args]

    return tp


def resolve_annotation(
    annotation: Any,
    globalns: dict[str, Any],
    localns: Optional[dict[str, Any]],
    cache: Optional[dict[str, Any]],
) -> Any:
    if annotation is None:
        return type(None)
    if isinstance(annotation, str):
        annotation = ForwardRef(annotation)

    locals = globalns if localns is None else localns
    if cache is None:
        cache = {}
    return evaluate_annotation(annotation, globalns, locals, cache)


BUILDING_DOCS: bool
try:
    BUILDING_DOCS = __sphinx_building__  # type: ignore
except NameError:
    BUILDING_DOCS = False
